/** @file launcher.c
 *
 *   The AQUOS Launcher
 *
 *     Copyright (c) 2005, 2006 Sharp Corporation
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/reboot.h>
#include <sys/time.h>
#include <sys/mount.h>

#include "../../launcher.h"

#define NELMS(array) (sizeof(array)/sizeof((array)[0]))


#if defined(DEBUG)
#define the_proctab the_proctab_debug
const static process_table_t the_proctab_debug[] =
{
  { Launch_Primary,    "./sample1" },
  { Launch_Essential,  "./sample2" },
  { Launch_Essential,  "./sample2" },
  { Launch_Repetitive, "./sample3" },
  { Sleep,             "3"         },
  { Launch_OneShot,    "./sample4" },
  { Wait,              ""          },
  { Launch_OneShot,    "./sample4" },
};
#endif /* defined(DEBUG) */


typedef enum msg_mode {
  QUIET_MODE,
  VERBOSE_MODE,
} msg_mode_t;

typedef enum launch_mode {
  NORMAL_MODE,
  DEBUG_MODE,
} launch_mode_t;

#if defined(UNHOSTED)
static msg_mode_t    the_message_mode = QUIET_MODE;
#else
static msg_mode_t    the_message_mode = VERBOSE_MODE;
#endif /* defined(UNHOSTED) */

static launch_mode_t the_launch_mode = NORMAL_MODE;

static char*         the_argv[MAX_NUM_OF_ARGS + 1];

extern int insmod_main(int argc, char* argv[]);

static void  print_timestamp(const char* string);
static void  print_status(int status, pid_t pid, const char* string);
static int   setup_args(process_table_t* table);
static int   minimize_thread_priority(void);
static void  launch(process_table_t* table);
static void  startup(void);
static void  maintain(void);
static void* thread_main(void* arg);
static void  restart_system();
static void  setup_mode(const char* cmdline);

static void print_timestamp(const char* string)
{
#ifdef ENABLE_DBGTIME
  char str[MAX_LENGTH_OF_COMMAND];
  snprintf(str, sizeof(str),"$DT launcher:%s",string);
  access(str, 0);
#else  
  if (the_message_mode != VERBOSE_MODE) {
    return;
  }
    
  struct timeval timeofday;
  gettimeofday(&timeofday, NULL);
  fprintf(stderr, "launcher: %s\n", string);
  fprintf(stderr, "launcher: %ld.%06ld\n",
          timeofday.tv_sec, timeofday.tv_usec);
#endif /* ENABLE_DBGTIME */
}

static void print_status(int status, pid_t pid, const char* string)
{
  if (the_message_mode != VERBOSE_MODE) {
    return;
  }
    
  fprintf(stderr, "Launcher: Process #%d (%s) ", pid, string);
  
  if (WIFEXITED(status)) {
    fprintf(stderr, "exited with status %d.\n", WEXITSTATUS(status));
  }
  else if (WIFSTOPPED(status)) {
    fprintf(stderr, "is stooped by Process #%d.\n", WSTOPSIG(status));
  }
  else if (WIFSIGNALED(status)) {
    int signal = WTERMSIG(status);
    switch (signal) {
      
    case SIGSEGV:
      fprintf(stderr, "caused Segmentation Fault.\n");
      break;

    case SIGBUS:
      fprintf(stderr, "caused Bus Error.\n");
      break;

    case SIGFPE:
      fprintf(stderr, "caused Floating Pointer Exception.\n");
      break;

    default:
      fprintf(stderr, "terminated by signal #%d.\n", signal);
    }
  }
  else {
    fprintf(stderr, "terminated by Unknown Reason (status = %d).\n", status);
  }
  fflush(stderr);
}

static int setup_args(process_table_t* table)
{
  static char  args_buffer[MAX_LENGTH_OF_COMMAND];
  char* p    = args_buffer;
  int   argc = 0;

  strncpy(args_buffer, table->string, sizeof(args_buffer) - 1);

  if ((the_launch_mode == DEBUG_MODE) && (table->dbgopts != NULL)) {
    int optstart = strlen(table->string);
    int optsize  = sizeof(args_buffer) - optstart;
    
    strncpy(args_buffer + optstart, table->dbgopts, optsize - 1);
  }
      
  args_buffer[sizeof(args_buffer) - 1] = '\0';

  while (argc < sizeof(the_argv) - 1) {
    while (isspace(*p))
      ++p;
    
    if (*p == '\0') {
      break;
    }
    
    the_argv[argc++] = p;
    
    while (!isspace(*p) && (*p != '\0'))
      ++p;
    
    if (*p == '\0') {
      break;
    }
    
    *p++ = '\0';
  }
  
  the_argv[argc] = NULL;
  return argc;
}

static int minimize_thread_priority(void)
{
#if defined(USE_THREAD)
  struct sched_param param;
  pthread_t thread = pthread_self();

  param.sched_priority = sched_get_priority_min(PROGRAMS_SCHED_POLICY);
  return pthread_setschedparam(thread, PROGRAMS_SCHED_POLICY, &param);
#else
  return 0;
#endif /* defined(USE_THREAD) */
}

static void launch(process_table_t* table)
{
  pid_t pid = 0;
  
  pid = fork();
  
  if (pid == 0) {
    if (table->action != Launch_Primary) {
      freopen("/dev/null", "r", stdin);
    }

    setup_args(table);
    
    minimize_thread_priority();
    
    execvp(the_argv[0], the_argv);
    perror("execvp");
    exit(EXIT_FAILURE);
  }
  else if (pid == -1) {
    perror("fork");
    exit(EXIT_FAILURE);
  }
  else {
    table->pid = pid;
  }
}

static void startup(void)
{
  int   i;
  pid_t prev_pid = 0;
    
  for (i = 0; i < NELMS(the_proctab); ++i) {

    print_timestamp(the_proctab[i].string);

    switch (the_proctab[i].action) {

    case Sleep:
      sleep(atoi(the_proctab[i].string));
      break;

    case Wait:
      waitpid(prev_pid, NULL, 0);
      break;

    case Insmod:
    {
      int argc = setup_args(&the_proctab[i]);
      insmod_main(argc, the_argv);
    }
    break;
    
    default:
      launch(&the_proctab[i]);
      prev_pid = the_proctab[i].pid;
    }
  }
}


static void maintain(void)
{
  int   i;
  pid_t pid;
  int   status;
  
  while ((pid = wait(&status)) != -1) {

    for (i = 0; i < NELMS(the_proctab); ++i) {
      if (pid == the_proctab[i].pid) {
        break;
      }
    }
    
    if (i == NELMS(the_proctab)) {
      continue;
    }
    
    print_status(status, pid, the_proctab[i].string);

    switch (the_proctab[i].action) {
      
    case Launch_Primary:
    case Launch_Essential:
      return;
      
    case Launch_OneShot:
      break;
      
    case Launch_Repetitive:
      launch(&the_proctab[i]);
      break;
      
    default:
      assert(0);
    }
  }
}

static void* thread_main(void* arg)
{
  startup();
  maintain();
  
  return NULL;
}

static void restart_system()
{
  if (the_message_mode == VERBOSE_MODE) {
    fprintf(stderr, "Rebooting...\n");
  }
      
  sync();
  kill(0, SIGKILL);
}

static void setup_mode(const char* cmdline)
{
  const char* postfix = DEBUG_PROGNAME_POSTFIX;
  const char* p       = rindex(cmdline, postfix[0]);

  if ((p != NULL) && (strncmp(p, postfix, strlen(postfix)) == 0)) {
    the_message_mode = VERBOSE_MODE;
    the_launch_mode  = DEBUG_MODE;
  }
}

int main(int argc, char* argv[])
{
  pid_t pid = fork();

  setup_mode(argv[0]);
    
  if (pid == 0) {

    /* Setup /proc filesystem */
    mount("proc", "/proc", "proc", 0, "defaults");

#if defined(USE_THREAD)
    {
      struct sched_param param;
      pthread_attr_t     attr;
      pthread_t          thread;
      int result;
      
      param.sched_priority = sched_get_priority_max(LAUNCHER_SCHED_POLICY);
      pthread_attr_init(&attr);
      pthread_attr_setschedpolicy (&attr, LAUNCHER_SCHED_POLICY);
      pthread_attr_setschedparam  (&attr, &param);
      
      result = pthread_create(&thread, &attr, thread_main, NULL);
      
      if (result == 0) {
        pthread_join(thread, NULL);
      }
    }
#else
    thread_main(NULL);
#endif /* defined(USE_THREAD) */
  }
  else {
    wait(NULL);
    restart_system();
  }

  return 0;
}
